local name = ...
local mod = LibStub("AceAddon-3.0"):NewAddon("BigBrother_Ready", "AceEvent-3.0", "AceTimer-3.0")

local L = LibStub("AceLocale-3.0"):GetLocale(name)

mod.version = GetAddOnMetadata(name, 'version')


--[===[@debug@
_G.BBReady = mod

local PREFIX = "|cff00ff00BB_Ready|r: "
local DevTools_Dump
DevTools_Dump = function(value)
	if not _G.DevTools_Dump then
		LoadAddOn("Blizzard_DebugTools")
	end
	DevTools_Dump = _G.DevTools_Dump
	DevTools_Dump(value)
end

local function dump(value)
	local orig = DEFAULT_CHAT_FRAME
	DEFAULT_CHAT_FRAME = ChatFrame4
	DevTools_Dump(value)
	DEFAULT_CHAT_FRAME = orig
end

local function mystrjoin(arg1, ...)
	local text = tostring(arg1)
	for i = 1, select('#', ...) do
		text = text .. tostring(select(i, ...))
	end
	return text
end

local function doDebug(arg1, ...)
	if not arg1 then return end
	if type(arg1) == 'table' then
		ChatFrame4:AddMessage(PREFIX .. "+++ table-dump")
		dump(arg1)
		ChatFrame4:AddMessage(PREFIX .. "--- end of table-dump")
		doDebug(...)
	else
		local text = mystrjoin(PREFIX, arg1, ...)
		ChatFrame4:AddMessage(text)
	end
end
--@end-debug@]===]

local function debug(...)
	--[===[@debug@
	doDebug(...)
	--@end-debug@]===]
end
mod.debug = debug

local foods = {
	87545, -- 90 Strength
	87554, -- 90 Dodge
	87555, -- 90 Parry
	87552, -- 90 Haste
	87551, -- 90 Crit
	87550, -- 90 hit
	87549, -- 90 Mastery
	87548, -- 90 Spirit
	87635, -- 90 Expertise
	87547, -- 90 Intellect
	87546, -- 90 Agility
}

local flasks = {
	94160, -- Flask of Flowing Water
	79469, -- Flask of Steelskin
	79472, -- Flask of Titanic Strength
	79470, -- Flask of the Draconic Mind
	79471, -- Flask of the Winds
	-- TODO Flask of Battle
}

local guardianElixirs = {
	79480, -- Elixir of Deep Earth
	79631, -- Prismatic Elixir
}

local battleElixirs = {
	79481, -- Impossible Accuracy
	79632, -- Mighty Speed
	79477, -- Elixir of the Cobra
	79635, -- Elixir of the Master
	79474, -- Elixir of the Naga
	79468, -- Ghost Elixir
}

local db
local timers = {}

local function checkForAnyBuff(unit, buffList)
	for i,v in ipairs(buffList) do
		local spellName = GetSpellInfo(v)
		local name, _, _, _, _, duration, expirationTime,_,_,_,spellid = UnitAura(unit, spellName)
		if name and spellid == v then
			if expirationTime then
				local remaining = expirationTime - GetTime()
				return remaining
			else
				return
			end
		end
	end
	return
end

local function checkForMissingElixir(unit)
	if db.onlyflasks then
		return L["No flask"]
	end
	local MIN_DURATION = db.threshold
	local guardianRemaining = checkForAnyBuff(unit, guardianElixirs)
	local battleRemaining = checkForAnyBuff(unit, battleElixirs)
	if guardianRemaining and battleRemaining then
		if guardianRemaining > MIN_DURATION and battleRemaining > MIN_DURATION then
			return
		elseif guardianRemaining < MIN_DURATION or battleRemaining < MIN_DURATION then
			return L["Expiring elixirs"]
		end
	end
	local onlybuff = guardianRemaining or battleRemaining
	if onlybuff then
		if onlybuff < MIN_DURATION then
			return L["Only one expiring elixir"]
		else
			return L["Only one elixir"]
		end
	else
		return L["No flask or elixir"]
	end
	return false
end

local function checkForMissingFlask(unit)
	local remaining = checkForAnyBuff(unit, flasks)
	if not remaining then
		return checkForMissingElixir(unit)
	elseif remaining < db.threshold then
		return remaining
	else
		return
	end
end

local function checkForMissingFood(unit)
	if not db.checkfood then
		return
	end
	local remaining = checkForAnyBuff(unit, foods)
	if remaining then
		if remaining < db.threshold then
			return remaining
		else
			return
		end
	elseif UnitAura(unit, GetSpellInfo(foods[1])) then
		return L["Low food buff"]
	else
		return L["No food buff"]
	end
end


local function broadCastToRaid(prefix, table)
	if #table <= 0 then
		return
	end
	local message = prefix .. ": " .. strjoin(", ", unpack(table))
	if db.channel == "SELF" then
		print("BBReady: ", message)
	elseif db.channel == "CUSTOM" then
		-- NYI
	else
		SendChatMessage(prefix .. ": " .. strjoin(", ", unpack(table)), db.channel)
	end
end
local TOKEN = "<BB_Ready>"
local outgoingFilter = function(self, event, msg, ...)
	if msg:find(TOKEN) then return true end
end
ChatFrame_AddMessageEventFilter("CHAT_MSG_WHISPER_INFORM", outgoingFilter)

local function broadCastWhisper(state, table)
	for _, v in ipairs(table) do
		SendChatMessage(("%s %s"):format(TOKEN, state), "WHISPER", nil, v)
	end
end

local function autoInitWithEmptyTable(self, key)
	local val = rawget(self, key)
	if not val then
		val = {}
		rawset(self, key, val)
	end
	return val
end
local mt = { __index = autoInitWithEmptyTable, }

local function getRelevantGroups()
	if db.groups.auto then
		local difficulty = GetRaidDifficulty()
		if difficulty%2 == 1 then
			return { true, true, }
		else
			return { true, true, true, true, true, }
		end
	else
		return CopyTable(db.groups)
	end
end

local function formatTime(seconds)
	local minutes = math.floor(seconds/60)
	local seconds = math.floor(seconds - minutes * 60)
	if seconds < 10 then
		seconds = "0" .. seconds
	end
	return ("%s:%s"):format(minutes, seconds)
end

local function iterateRaidMembers()
	local i = 0
	local n = GetNumRaidMembers()
	return function()
			i = i + 1
			if i <= n then
				local unit = "raid" .. i
				local name = UnitName(unit)
				local _, _, group = GetRaidRosterInfo(i)
				return unit, name, group
			end
		end
end

function mod:READY_CHECK()
	debug("ready_check")
	local buffStates = setmetatable({}, mt)
	local groups = getRelevantGroups()
	for unit, name, group in iterateRaidMembers() do
		if not groups[group] then
			-- nothing
		elseif not UnitIsVisible(unit) then
			table.insert(buffStates[L["Out of range"]], name)
		else
			local state = checkForMissingFlask(unit)
			if type(state) == 'number' then
				table.insert(buffStates[L["Expiring flask"]], ("%s (%s)"):format(name, formatTime(state)))
			elseif state then
				table.insert(buffStates[state], name)
			end
			local foodstate = checkForMissingFood(unit)
			if type(foodstate) == 'number' then
				table.insert(buffStates[L["Expiring food buff"]], ("%s (%s)"):format(name, formatTime(foodstate)))
			elseif foodstate then
				table.insert(buffStates[foodstate], name)
			end
		end
	end
	if db.channel == "RAID" and (not mod.speaker or mod.speaker == playerName) then
		return
	end
	for k,v in pairs(buffStates) do
		broadCastToRaid(k, v)
	end
	if db.naggingWhispers then
		wipe(buffStates[L["Out of range"]])
		for k,v in pairs(buffStates) do
			broadCastWhisper(k, v)
		end
	end
	-- cleanup the mess
	wipe(buffStates)
	buffStates = nil
	wipe(groups)
	groups = nil
end

local function checkForBoss()
	-- TODO implement
	return false
end

function mod:PLAYER_REGEN_ENABLED()
	if checkForBoss() then
		self:ScheduleTimer("READY_CHECK", 10)
	end
end

local dbDefaults = {
	profile = {
		suspend = false,
		naggingWhispers = false,
		threshold = 10*60, -- 10 minutes
		groups = { auto = true },
		channel = "SELF",
		flaskonly = false,
		checkfood = true,
	}
}
for i = 1,8 do
	dbDefaults.profile.groups[i] = true
end

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("BigBrother_ReadyDB", dbDefaults)
	db = self.db.profile
end

local ldbdata
function mod:OnEnable()
	debug("enabling")
	if db.suspend then
		ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-NotReady]]
		ldbdata.text = "Off"
	else
		self:Resume()
	end
end

function mod:Resume()
	self:RegisterEvent("CHAT_MSG_ADDON")
	self:RegisterEvent("RAID_ROSTER_UPDATE", "UpdateMembers")
	self:RegisterEvent("READY_CHECK")
	self:Resync()
	ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-Ready]]
	if db.naggingWhispers then
		ldbdata.text = "On|cffff00ff/w|r"
	else
		ldbdata.text = "On"
	end
end

function mod:Suspend()
	self:UnregisterAllEvents()
	self:CancelAllTimers()
	self:Resync()
	ldbdata.icon = [[Interface\RAIDFRAME\ReadyCheck-NotReady]]
	ldbdata.text = "Off"
	mod.speaker = nil
end

function mod:OnDisable()
	self:Suspend()
end

local function toggleState()
	db.suspend = not db.suspend
	if db.suspend then
		mod:Suspend()
	else
		mod:Resume()
	end
end

ldbdata = {
	type = "launcher",
	OnClick = function(_, msg)
			if msg == "LeftButton" then
				if IsShiftKeyDown() then
					mod:READY_CHECK()
				else
					toggleState()
				end
			elseif msg == "RightButton" then
				mod:ShowConfig()
			end
		end,
	OnTooltipShow = function(tt)
			tt:AddLine("BigBrother - Ready")
			if not db.suspend then
				tt:AddLine("is |cff00ff00on|r")
			else
				tt:AddLine("is |cffff0000off|r")
			end
			if db.naggingWhispers then
				tt:AddLine("Additionally I will nag people with |cffff00ffwhipsers|r")
			end
			tt:AddLine("|cffffff00Click|r switch on/off")
			tt:AddLine("|cffffff00Shift-Click|r force buff-check")
			tt:AddLine("|cffffff00Right-Click|r open options")
			if mod.speaker then
				tt:AddLine("|cff00ff00 Speaker:|r " .. mod.speaker)
			end
		end,
	label = "BigBrother_Ready",
}

LibStub("LibDataBroker-1.1"):NewDataObject("BigBrother_Ready", ldbdata)
mod.ldbdata = ldbdata
-- sync-stuff
local candidates = {}
local ADDON_MSG_PREFIX = "BBReady"
local playerName = UnitName("player")

function mod:Resync()
	debug("resyncing")
	SendAddonMessage(ADDON_MSG_PREFIX, "resync", "RAID")
end

local function monitorSpeaker(self)
	if not UnitIsConnected(mod.speaker) then
		debug("speaker went offline")
		mod:CheckGroup()
		mod:CancelTimer(timers.monitorSpeaker, true)
		timers.monitorSpeaker = nil
	end
end


function mod:ChooseCandidate()
	debug("timer expired")
	if #candidates > 0 then
		table.sort(candidates)
		debug(candidates)
		mod.speaker = candidates[1]
		if playerName == mod.speaker then
			debug("yesss, it's me")
		else
			debug(":( it's not me")
			timers.monitorSpeaker = self:ScheduleRepeatingTimer(monitorSpeaker, 1)
		end
	else
		debug("no candidates found")
	end
	debug(mod.speaker, " has been chosen")
	if mod.speaker ~= playerName then
		ldbdata.text = "|c88888800On|r"
	else
		ldbdata.text = "On"
	end
	wipe(candidates)
	bestCandidateRating = 0
	timers.ChooseCandidate = nil
end

local function getVersionValue()
	local mversion, alpha = ("-"):split(mod.version)
	local major, minor, rev = ("."):split(mversion)
	if not tonumber(major) then
		return 99999
	end
	local value = major * 10000 + minor *1000 + rev
	if alpha then
		value = value + alpha
	end
	return value
end

-- ping-value is calculated like: version > raidrank > naggingWhispers
local function getPingValue()
	local version = getVersionValue()
	local raidrank = (IsRaidLeader() and 2) or (IsRaidOfficer() and 1) or 0
	local nagValue = db.naggingWhispers and 1 or 0
	return version * 100 + raidrank * 10 + nagValue
end

local function sendPing()
	if db.channel ~= "RAID" then
		return
	end
	debug("ping the raid")
	local message = getPingValue()
	SendAddonMessage(ADDON_MSG_PREFIX, message, "RAID")
	timers.pingTimer = nil
end

local inRaid = false
function mod:UpdateMembers()
	local newState = not not UnitInRaid("player")
	debug(newState, "<->", inRaid)
	if newState ~= inRaid then
		debug("changed RAID-state")
		inRaid = newState
		self:Resync()
		return
	end
	if not UnitInRaid(mod.speaker) then
		debug("the SPEAKER left the raid")
		self:CheckGroup()
	end
end

local bestCandidateRating = 0

function mod:CheckGroup(arg, ...)
	bestCandidateRating = 0
	wipe(candidates)
	if timers.ChooseCandidate then
		debug("reschedule")
		self:CancelTimer(timers.ChooseCandidate, true)
	end
	debug("schedule timer")
	timers.ChooseCandidate = self:ScheduleTimer("ChooseCandidate", 10)
	self:CancelTimer(timers.sendping, true)
	timers.sendping = self:ScheduleTimer(sendPing, 2)
end

function mod:CHAT_MSG_ADDON(event, prefix, message, _, sender)
	--debug(prefix, " ", message, " ", sender)
	if prefix == ADDON_MSG_PREFIX then
		debug("message: ", message, " (", sender, ")")
		if message == "resync" then
			debug("recieved RESYNC from ", sender)
			self:CheckGroup()
		end
		local rating = tonumber(message)
		debug(tostring(rating), " <-> ", bestCandidateRating)
		if not rating or bestCandidateRating > rating then
			return
		elseif bestCandidateRating < rating then
			wipe(candidates)
			bestCandidateRating = rating
		end
		table.insert(candidates, sender)
	end
end

local channelChoices = {
	RAID = "Raid",
	YELL = "Yell",
	SAY = "Say",
	SELF = "Self",
	-- CUSTOM = "Custom",
}
local function getChannel()
	return db.channel
end
local function setChannel(info, value)
	db.channel = value
	mod:Resync()
end

local groupChoices = {}
for i=1,8 do
	groupChoices[i] = i
end

local function getIncludeGroup(info, key)
	return db.groups[key]
end

local function setIncludeGroup(info, key, value)
	db.groups[key] = value
end

local function setOption(info, value)
	local key = info[#info]
	debug("setting ", key, " -> ", value)
	db[key] = value
end
local function getOption(info)
	local key = info[#info]
	return db[key]
end

local options = {
	name = "BigBrother - Ready",
	desc = "BigBrother - Ready",
	type = 'group',
	icon = [[Interface\GROUPFRAME\UI-Group-LeaderIcon]],
	args = {
		channel = {
			name = L["Channel"],
			desc = L["Which channel to broadcast to"],
			type = 'select',
			values = channelChoices,
			set = function(...)
				setOption(...)
				mod:Resync()
			end,
			get = getOption,
			order = 1,
		},
		naggingWhispers = {
			name = L["Nagging whispers"],
			desc = L["Additionally send whispers to players that are missing buffs"],
			type = 'toggle',
			get = getOption,
			set = function(info, value)
					db.naggingWhispers = value
					mod:Resync()
				end,
			order = 2,
		},
		threshold = {
			name = L["Expiration Threshold"],
			desc = L["Report Buffs that expire in x minutes"],
			type = 'range',
			min = 0,
			max = 60,
			step = 1,
			set = function(info, value) db.threshold = value*60 end,
			get = function(info) return db.threshold/60 end,
			order = 3,
		},
		flaskonly = {
			name = L["Flasks only"],
			desc = L["Only count flasks as valid buff, not elixirs"],
			type = 'toggle',
			get = getOption,
			set = setOption,
			order = 4,
		},
		checkfood = {
			name = L["Check food"],
			desc = L["Also check the players for food buffs"],
			type = 'toggle',
			get = getOption,
			set = setOption,
			order = 5,
		},
		includeGroups = {
			name = L["Groups"],
			desc = L["Only include these groups"],
			type = 'group',
			inline = true,
			args = {
				auto = {
					type = 'toggle',
					name = L["Auto"],
					desc = L["Determine Groups automatically"],
					get = function(info) return db.groups.auto end,
					set = function(info, value) db.groups.auto = value end,
				},
				includeGroups = {
					type = 'multiselect',
					name = L["Groups"],
					desc = L["Specifiy groups to include in the check"],
					values = groupChoices,
					get = getIncludeGroup,
					set = setIncludeGroup,
					disabled = function() return db.groups.auto end,
				},
			},
		},
	},
}

local AceConfig = LibStub("AceConfig-3.0")
local AceConfigDialog = LibStub("AceConfigDialog-3.0")
AceConfig:RegisterOptionsTable("BBReady", options)
AceConfigDialog:SetDefaultSize("BBReady", 500, 350)
function mod:ShowConfig()
	AceConfigDialog:Open("BBReady")
end